from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Label


class CheckboxContainer(Widget):

    DEFAULT_CSS = """
    CheckboxContainer {
        layout: grid;
        grid-size: 2 1;
        grid-columns: 3 1fr;
        grid-gutter: 0 1;
        border: round $border-blurred;
        padding: 0 1;

        .checkbox-status {
            color: $foreground-muted;
        }

        &.checked {
            border: round $border;

            .checkbox-status {
                color: $text-success;
            }
        }
    }
    """

    __STATUS_CHECKED = "(@)"
    __STATUS_UNCHECKED = "( )"
    __CLASS_CHECKED = "checked"

    checked: reactive[bool] = reactive(False)

    content: Widget

    def __init__(self, content: Widget) -> None:
        super().__init__()
        self.content = content

    def compose(self) -> ComposeResult:
        yield Label(classes="checkbox-status", markup=False)
        yield self.content

    def watch_checked(self, checked: bool) -> None:
        label = self.query_exactly_one(".checkbox-status", Label)
        label.update(self.__STATUS_CHECKED if checked else self.__STATUS_UNCHECKED)
        self.set_class(checked, self.__CLASS_CHECKED)

    def on_click(self) -> None:
        self.checked = not self.checked



from typing import Literal, TypedDict, assert_never, get_args

type BasicCardColor = Literal["red", "green", "blue", "yellow"]
basic_card_colors: tuple[BasicCardColor, ...] = get_args(BasicCardColor.__value__)


class NumberCard(TypedDict):
    id: str
    color: BasicCardColor
    type: Literal["number"]
    number: int


type FunctionCardEffect = Literal["skip", "+2", "reverse"]
function_card_effects: tuple[FunctionCardEffect, ...] = get_args(
    FunctionCardEffect.__value__
)


class FunctionCard(TypedDict):
    id: str
    color: BasicCardColor
    type: Literal["function"]
    effect: FunctionCardEffect


type WildCardEffect = Literal["+4", "color"]
wild_card_effects: tuple[WildCardEffect, ...] = get_args(WildCardEffect.__value__)
type WildCardColor = Literal["black"]
wild_card_color: WildCardColor = "black"


class WildCard(TypedDict):
    id: str
    color: WildCardColor
    type: Literal["wild"]
    effect: WildCardEffect


type Card = NumberCard | FunctionCard | WildCard
type Deck = list[Card]


def format_card(card: Card) -> str:
    match card["type"]:
        case "number":
            return card["color"].title() + "-" + str(card["number"])
        case "function":
            return card["color"].title() + "-" + card["effect"].title()
        case "wild":
            return card["effect"].title()
        case _ as card_type:
            assert_never(card_type)



from typing import cast

from flask import Blueprint, request
from flask.typing import ResponseReturnValue

from tuno.server.exceptions import InvalidRequestBodyException
from tuno.server.utils.checkers import check_player_name
from tuno.server.utils.Logger import Logger

logger = Logger(__name__)


def setup(blueprint: Blueprint) -> None:

    @blueprint.put("/rules")
    def set_rules() -> ResponseReturnValue:

        player_name = request.args.get("player_name", "")
        if player_name:
            check_player_name(player_name)
        else:
            player_name = "(anonymous)"

        modified_rules = request.json
        if not isinstance(modified_rules, dict):
            raise InvalidRequestBodyException(
                f"expected a dict of modified rules, got {modified_rules!r}"
            )

        from tuno.server.models.Game import game

        game.update_rules(
            cast(dict[str, object], modified_rules),
            operator_name=player_name,
            operator_is_player=True,
        )

        return ("Rule updated.", 200)



from collections.abc import Iterable

from tuno.shared.constraints import MIN_PLAYER_CAPACITY
from tuno.shared.deck import BasicCardColor, Card


class ApiException(Exception):

    http_code: int
    message: str

    def __init__(self, http_code: int, message: str) -> None:
        super().__init__(message)
        self.http_code = http_code
        self.message = message


class InvalidRequestBodyException(ApiException):
    def __init__(self, message: str) -> None:
        super().__init__(
            400,
            message,
        )


class InvalidPlayerNameException(ApiException):
    def __init__(self, player_name: str) -> None:
        super().__init__(
            400,
            f"Invalid player name: {player_name}",
        )


class PlayerNotFoundException(ApiException):
    def __init__(self, player_name: str) -> None:
        super().__init__(
            400,
            f"Player not found: {player_name}",
        )


class NewPlayerToStartedGameException(ApiException):
    def __init__(self, player_name: str) -> None:
        super().__init__(
            403,
            f"Cannot add new player to a started game. ({player_name})",
        )


class GameAlreadyStartedException(ApiException):
    def __init__(self) -> None:
        super().__init__(
            400,
            "The game has already started.",
        )


class RuleUpdateOnStartedGameException(ApiException):
    def __init__(self) -> None:
        super().__init__(
            400,
            "Cannot update rules on a started game.",
        )


class NotEnoughPlayersException(ApiException):
    def __init__(self) -> None:
        super().__init__(
            400,
            "Not enough players to start the game. "
            f"(At least {MIN_PLAYER_CAPACITY} players are required.)",
        )


class GameNotStartedException(ApiException):
    def __init__(self) -> None:
        super().__init__(
            400,
            "The game has not started yet.",
        )


class InvalidLeadCardInfoException(ApiException):
    def __init__(
        self,
        lead_card: Card | None,
        lead_color: BasicCardColor | None,
    ) -> None:
        super().__init__(
            400,
            f"Invalid lead card info: {lead_card!r}, {lead_color!r}",
        )


class CardIdsNotFoundException(ApiException):
    def __init__(self, card_ids: Iterable[str]) -> None:
        super().__init__(
            400,
            f"Card id(s) not found: {', '.join(card_ids)}",
        )


class NotCurrentPlayerException(ApiException):
    def __init__(self, current_player_name: str) -> None:
        super().__init__(
            400,
            "Sorry, you are not the current player "
            f"(which is {current_player_name}).",
        )



from os import environ
from typing import cast

from textual.app import ComposeResult
from textual.screen import Screen
from textual.validation import Regex
from textual.widgets import Button, Footer, Header, Input

from tuno.client.config import (
    ENV_KEY_CONNECTION,
    ENV_KEY_PLAYER_NAME,
    ENV_KEY_SERVER_ADDRESS,
)
from tuno.shared.constraints import PLAYER_NAME_PATTERN

player_name_validator = Regex(PLAYER_NAME_PATTERN)


class ConnectScreen(Screen[object]):

    TITLE = "UNO"
    CSS_PATH = "styles.tcss"
    BINDINGS = [
        ("ctrl+j", "submit", "Submit"),
    ]

    def compose(self) -> ComposeResult:

        default_player_name = ""
        default_server_address = ""
        default_connection_info = environ.get(ENV_KEY_CONNECTION, "")
        if default_connection_info and "@" in default_connection_info:
            default_player_name, _, default_server_address = (
                default_connection_info.partition("@")
            )

        default_player_name = environ.get(
            ENV_KEY_PLAYER_NAME,
            default_player_name,
        )
        if not player_name_validator.validate(default_player_name).is_valid:
            default_player_name = ""
        if default_player_name:
            self.log.debug("Found default player name:", default_player_name)

        default_server_address = environ.get(
            ENV_KEY_SERVER_ADDRESS,
            default_server_address,
        )
        if default_server_address:
            self.log.debug(
                "Found default server address:",
                default_server_address,
            )

        player_name_pattern_display = PLAYER_NAME_PATTERN.pattern
        input_player_name = Input(
            id="player_name",
            classes="form-item",
            placeholder=f"Your name ({player_name_pattern_display})",
            validators=[player_name_validator],
            value=default_player_name,
        )
        input_player_name.border_title = "Player Name"

        input_server_address = Input(
            id="server_address",
            classes="form-item",
            placeholder="HOST:PORT",
            value=default_server_address,
        )
        input_server_address.border_title = "Server Address"

        yield Header(show_clock=True)
        yield input_player_name
        yield input_server_address
        yield Button(
            "JOIN",
            id="submit",
            classes="form-item",
            variant="primary",
            action="submit",
            disabled=True,
        )
        yield Footer()

    def action_submit(self) -> None:

        from tuno.client.UnoApp import UnoApp

        app = cast(UnoApp, self.app)
        assert isinstance(app, UnoApp)

        input_player_name = self.query_exactly_one("#player_name", Input)
        player_name = input_player_name.value
        if not player_name or not input_player_name.is_valid:
            self.log.warning(f"Invalid player name: {player_name}")
            return

        input_server_address = self.query_exactly_one("#server_address", Input)
        server_address = input_server_address.value
        if not server_address or not input_server_address.is_valid:
            self.log.warning(f"Invalid server address: {server_address}")
            return

        submit_button = self.query_exactly_one("#submit", Button)

        input_player_name.disabled = True
        input_server_address.disabled = True
        submit_button.disabled = True
        submit_button.loading = True

        def on_success() -> None:
            app.clear_notifications()
            app.notify("Connected to game server!")

        def on_failure(error_message: str) -> None:
            input_player_name.disabled = False
            input_server_address.disabled = False
            submit_button.disabled = False
            submit_button.loading = False

            app.notify_error(
                error_message,
                title="Connection Error",
            )

        assert app.client is not None
        app.client.connect(
            server_address=server_address,
            player_name=player_name,
            on_success=on_success,
            on_failure=on_failure,
        )

    def on_input_changed(self, event: Input.Changed) -> None:
        if not event.value or not event.input.is_valid:
            event.input.add_class("invalid")
        else:
            event.input.remove_class("invalid")
        self.query_exactly_one("#submit", Button).disabled = not all(
            (input.value and input.is_valid) for input in self.query(Input)
        )



from typing import TYPE_CHECKING

from tuno.shared.sse_events import NotificationEvent

if TYPE_CHECKING:
    from tuno.client.UnoApp import UnoApp


def handler(parsed_data: NotificationEvent.DataType, app: "UnoApp") -> None:
    app.notify(
        parsed_data["message"],
        title=parsed_data["title"],
    )



from flask import Blueprint, request
from flask.typing import ResponseReturnValue

from tuno.server.utils.checkers import check_player_name
from tuno.server.utils.Logger import Logger

logger = Logger(__name__)


def setup(blueprint: Blueprint) -> None:

    @blueprint.put("/stop")
    def stop() -> ResponseReturnValue:

        player_name = request.args.get("player_name", "")
        if player_name:
            check_player_name(player_name)
        else:
            player_name = "(anonymous)"

        from tuno.server.models.Game import game

        game.stop(player_name, operator_is_player=True)

        return ("Stopped!", 200)



from typing import Final

from textual.app import ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Label

from tuno.client.components.CardColorLabel import CardColorLabel
from tuno.client.components.CardLabel import CardLabel
from tuno.shared.sse_events import GameStateEvent


class Sidebar(VerticalScroll):

    __CLASS_GAME_STATUS_STARTED: Final = "started"
    __CLASS_GAME_STATUS_PENDING: Final = "pending"
    __CLASS_CURRENT_PLAYER_ACTIVE: Final = "active"
    __CLASS_CURRENT_PLAYER_WAITING: Final = "waiting"

    BORDER_TITLE = "Game Info"
    DEFAULT_CSS = """
    Sidebar {
        row-span: 2;
        height: 100%;
        padding: 1 2;

        .sidebar-info-section {
            width: 100%;
            height: 3;
            align-horizontal: center;
            content-align-horizontal: center;
            color: $foreground-muted;
            border-title-color: $foreground;
            border-title-align: center;
            border: solid $surface;

            .sidebar-info-split {
                color: $text-muted 20%;
                margin: 0 1;
            }
        }

        #sidebar-info-status {
            &.started {
                color: $text-success;
            }
            &.pending {
                color: $text-warning;
            }
        }

        #sidebar-info-current-player {
            color: $foreground-muted;

            &.active {
                color: $text-success;
            }
            &.waiting {
                color: $text-accent;
            }
        }

        #sidebar-info-pile-size {
            #sidebar-info-pile-size-draw {
                color: $text-success;
            }

            #sidebar-info-pile-size-discard {
                color: $text-error;
            }
        }

        #sidebar-info-counters {
            #sidebar-info-draw-counter, #sidebar-info-skip-counter {
                color: $foreground;
            }
        }

        #sidebar-info-lead-color-container, #sidebar-info-lead-card-container {
            padding: 0 1;
        }
    }
    """

    game_state: reactive[GameStateEvent.DataType | None] = reactive(None)

    def compose(self) -> ComposeResult:

        label_status = Label(
            "...",
            id="sidebar-info-status",
            classes="sidebar-info-section",
        )
        label_status.border_title = "Game Status"
        yield label_status

        label_capacity = Label(
            "-/-",
            id="sidebar-info-capacity",
            classes="sidebar-info-section",
        )
        label_capacity.border_title = "Capacity"
        yield label_capacity

        label_direction = Label(
            "-/-",
            id="sidebar-info-direction",
            classes="sidebar-info-section",
        )
        label_direction.border_title = "Direction"
        yield label_direction

        label_current_player = Label(
            "-/-",
            id="sidebar-info-current-player",
            classes="sidebar-info-section",
        )
        label_current_player.border_title = "Current"
        yield label_current_player

        label_draw_pile_size = Label(
            "-",
            id="sidebar-info-pile-size-draw",
        )
        label_draw_pile_size.tooltip = "Draw pile."
        label_discard_pile_size = Label(
            "-",
            id="sidebar-info-pile-size-discard",
        )
        label_discard_pile_size.tooltip = "Discard pile."
        container_pile_size = Horizontal(
            label_draw_pile_size,
            Label("|", classes="sidebar-info-split"),
            label_discard_pile_size,
            id="sidebar-info-pile-size",
            classes="sidebar-info-section",
        )
        container_pile_size.border_title = "Pile Size"
        yield container_pile_size

        label_draw_counter = Label(
            "-",
            id="sidebar-info-draw-counter",
        )
        label_draw_counter.tooltip = "Draw counter."
        label_skip_counter = Label(
            "-",
            id="sidebar-info-skip-counter",
        )
        label_skip_counter.tooltip = "Skip counter."
        container_counters = Horizontal(
            label_draw_counter,
            Label("|", classes="sidebar-info-split"),
            label_skip_counter,
            id="sidebar-info-counters",
            classes="sidebar-info-section",
        )
        container_counters.border_title = "Draw/Skip"
        yield container_counters

        widget_lead_color = Widget(
            CardColorLabel(id="sidebar-info-lead-color"),
            id="sidebar-info-lead-color-container",
            classes="sidebar-info-section",
        )
        widget_lead_color.border_title = "Lead Color"
        yield widget_lead_color

        widget_lead_card = Widget(
            CardLabel(id="sidebar-info-lead-card"),
            id="sidebar-info-lead-card-container",
            classes="sidebar-info-section",
        )
        widget_lead_card.border_title = "Lead Card"
        yield widget_lead_card

    def watch_game_state(
        self,
        game_state: GameStateEvent.DataType | None,
    ) -> None:

        from tuno.client.UnoApp import UnoApp

        app = self.app
        assert isinstance(app, UnoApp)
        assert app.client is not None

        # -- Game Status --
        label_game_status = self.query_exactly_one(
            "#sidebar-info-status",
            Label,
        )
        if game_state:
            if game_state["started"]:
                game_status_display = "Started"
                label_game_status.add_class(self.__CLASS_GAME_STATUS_STARTED)
                label_game_status.remove_class(self.__CLASS_GAME_STATUS_PENDING)
            else:
                game_status_display = "Pending"
                label_game_status.remove_class(self.__CLASS_GAME_STATUS_STARTED)
                label_game_status.add_class(self.__CLASS_GAME_STATUS_PENDING)
        else:
            game_status_display = "..."
            label_game_status.remove_class(
                self.__CLASS_GAME_STATUS_STARTED,
                self.__CLASS_GAME_STATUS_PENDING,
            )
        label_game_status.update(game_status_display)

        # -- Player Capacity --
        label_player_capacity = self.query_exactly_one(
            "#sidebar-info-capacity",
            Label,
        )
        if game_state:
            player_count = len(game_state["players"])
            player_capacity = game_state["rules"]["player_capacity"]
            label_player_capacity.update(f"{player_count}/{player_capacity}")
        else:
            label_player_capacity.update("-/-")

        # -- Direction --
        label_direction = self.query_exactly_one(
            "#sidebar-info-direction",
            Label,
        )
        if game_state and game_state["started"]:
            direction = f"{game_state['direction']:+d}"
            label_direction.update(direction)
        else:
            label_direction.update("N/A")

        # -- Current Player --
        label_current_player = self.query_exactly_one(
            "#sidebar-info-current-player",
            Label,
        )
        current_player_name = "N/A"
        if game_state and game_state["started"]:
            current_player_index = game_state["current_player_index"]
            players = game_state["players"]
            if 0 <= current_player_index < len(players):
                current_player_name = players[current_player_index]["name"]
            else:
                self.log.error(
                    "Invalid value for `current_player_index`:",
                    current_player_index,
                )
            if current_player_name == app.client.player_name:
                label_current_player.add_class(self.__CLASS_CURRENT_PLAYER_ACTIVE)
                label_current_player.remove_class(self.__CLASS_CURRENT_PLAYER_WAITING)
            else:
                label_current_player.remove_class(self.__CLASS_CURRENT_PLAYER_ACTIVE)
                label_current_player.add_class(self.__CLASS_CURRENT_PLAYER_WAITING)
        else:
            label_current_player.remove_class(
                self.__CLASS_CURRENT_PLAYER_ACTIVE,
                self.__CLASS_CURRENT_PLAYER_WAITING,
            )
        label_current_player.update(current_player_name)

        # -- Pile Size --
        label_draw_pile_size = self.query_exactly_one(
            f"#sidebar-info-pile-size-draw",
            Label,
        )
        label_discard_pile_size = self.query_exactly_one(
            f"#sidebar-info-pile-size-discard",
            Label,
        )
        if game_state and game_state["started"]:
            label_draw_pile_size.update(str(game_state["draw_pile_size"]))
            label_discard_pile_size.update(str(game_state["discard_pile_size"]))
        else:
            label_draw_pile_size.update("-")
            label_discard_pile_size.update("-")

        # -- Draw/Skip Counters --
        label_draw_count = self.query_exactly_one(
            f"#sidebar-info-draw-counter",
            Label,
        )
        label_skip_count = self.query_exactly_one(
            f"#sidebar-info-skip-counter",
            Label,
        )
        if game_state and game_state["started"]:
            label_draw_count.update(str(game_state["draw_counter"]))
            label_skip_count.update(str(game_state["skip_counter"]))
        else:
            label_draw_count.update("-")
            label_skip_count.update("-")

        # -- Lead Color --
        widget_lead_color = self.query_exactly_one(
            f"#sidebar-info-lead-color",
            CardColorLabel,
        )
        widget_lead_color.data = game_state and game_state["lead_color"]

        # -- Lead Card --
        widget_lead_card = self.query_exactly_one(
            f"#sidebar-info-lead-card",
            CardLabel,
        )
        widget_lead_card.data = game_state and game_state["lead_card"]



from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, Self

from requests import RequestException

if TYPE_CHECKING:
    from tuno.client.UnoApp import UnoApp


class ApiContext(AbstractContextManager["ApiContext"]):

    error_title: str
    app: "UnoApp"

    def __init__(self, error_title: str, *, app: "UnoApp") -> None:
        self.error_title = error_title
        self.app = app

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exception_type: type[BaseException] | None,
        exception: BaseException | None,
        traceback: object,
    ) -> bool:

        if exception:

            if isinstance(exception, RequestException):

                error_message = "RequestException caught in ApiContext:"
                self.app.log.warning(
                    error_message,
                    exception,
                )

                if exception.response is not None:
                    error_message = exception.response.text
                else:
                    error_message = f"{error_message} {exception!r}"

                self.app.notify_error(
                    error_message,
                    title=self.error_title,
                )

                return True  # stop exception propagation

            else:

                error_message = "Unknown exception caught in ApiContext:"
                self.app.log.error(
                    error_message,
                    exception,
                )
                self.app.notify_error(
                    f"{error_message} {exception!r}",
                    title=self.error_title,
                )

        return False  #  propagate the exception



from collections.abc import Mapping, Sequence
from random import shuffle
from threading import RLock, Thread
from time import monotonic
from typing import TYPE_CHECKING, Literal, assert_never

from tuno.server.config import (
    GAME_WATCHER_INTERVAL,
    GAME_WATCHER_SKIP_THRESHOLD,
    PLAYER_TIMEOUT,
)
from tuno.server.exceptions import (
    ApiException,
    GameAlreadyStartedException,
    GameNotStartedException,
    InvalidLeadCardInfoException,
    NewPlayerToStartedGameException,
    NotCurrentPlayerException,
    NotEnoughPlayersException,
    PlayerNotFoundException,
    RuleUpdateOnStartedGameException,
)
from tuno.server.utils.create_deck import create_deck
from tuno.server.utils.format_optional_operator import format_optional_operator
from tuno.server.utils.Logger import Logger
from tuno.shared.check_play import check_play
from tuno.shared.constraints import (
    DEFAULT_INITIAL_HAND_SIZE,
    DEFAULT_PLAYER_CAPACITY,
    MIN_PLAYER_CAPACITY,
)
from tuno.shared.deck import BasicCardColor, Card, Deck, format_card
from tuno.shared.loop import loop
from tuno.shared.rules import GameRules, check_rule_update
from tuno.shared.sse_events import (
    EndOfConnectionEvent,
    GameStateEvent,
    NotificationEvent,
    ServerSentEvent,
)

from .Player import Player


class Game:

    tag: str = "default"  # TODO: maybe multiple games in one server someday
    lock: RLock
    watcher_thread: Thread

    __players: list[Player]
    __started: bool
    __rules: GameRules
    __draw_pile: Deck
    __discard_pile: Deck
    __current_player_index: int
    __direction: Literal[-1, 1]
    __lead_card: Card | None
    __lead_color: BasicCardColor | None
    __draw_counter: int
    __skip_counter: int
    __logger: Logger

    def __init__(self) -> None:

        self.__players = []
        self.__started = False
        self.__rules = GameRules(
            player_capacity=DEFAULT_PLAYER_CAPACITY,
            initial_hand_size=DEFAULT_INITIAL_HAND_SIZE,
            shuffle_players=True,
            any_last_play=True,
        )
        self.__draw_pile = []
        self.__discard_pile = []
        self.__current_player_index = -1
        self.__direction = 1
        self.__lead_card = None
        self.__lead_color = None
        self.__draw_counter = 0
        self.__skip_counter = 0
        self.lock = RLock()
        self.__logger = Logger(f"{__name__}#{self.tag}")
        self.watcher_thread = Thread(target=self.watcher_loop, daemon=True)

        self.__logger.debug(f"game#{self.tag} created")

    @property
    def started(self) -> bool:
        return self.__started

    def broadcast(self, event: ServerSentEvent) -> None:
        with self.lock:
            for player in self.__players:
                player.message_queue.put(event)

    def get_game_state_event(self) -> GameStateEvent:
        with self.lock:
            started = self.__started
            return GameStateEvent(
                GameStateEvent.DataType(
                    started=started,
                    rules=self.__rules,
                    draw_pile_size=(len(self.__draw_pile) if started else -1),
                    discard_pile_size=(len(self.__discard_pile) if started else -1),
                    players=[
                        GameStateEvent.PlayerDataType(
                            name=player.name,
                            connected=player.connected,
                            card_count=(
                                len(player.cards) if started else player.last_result
                            ),
                        )
                        for player in self.__players
                    ],
                    current_player_index=self.__current_player_index,
                    direction=self.__direction,
                    lead_card=self.__lead_card,
                    lead_color=self.__lead_color,
                    draw_counter=self.__draw_counter,
                    skip_counter=self.__skip_counter,
                )
            )

    def update_rules(
        self,
        modified_rules: Mapping[str, object],
        *,
        operator_name: str | None,
        operator_is_player: bool,
    ) -> None:

        with self.lock:

            if self.started:
                debug_message = format_optional_operator(
                    "Rejected rule update",
                    operator_name,
                    is_player=operator_is_player,
                )
                self.__logger.debug(debug_message)
                raise RuleUpdateOnStartedGameException()

            rules = self.__rules.copy()
            for key, value in modified_rules.items():
                check_rule_update(key, value)
                rules[key] = value  # type: ignore[literal-required]
            self.__rules = rules

            message = format_optional_operator(
                "Game rules updated",
                operator_name,
                is_player=operator_is_player,
            )
            self.__logger.info(message)
            self.broadcast(
                NotificationEvent(
                    NotificationEvent.DataType(
                        title="Rules Updated",
                        message=message,
                    )
                )
            )

            if "player_capacity" in modified_rules:
                new_capacity = self.__rules["player_capacity"]
                if len(self.__players) > new_capacity:
                    while len(self.__players) > new_capacity:
                        excess_player = self.__players.pop()
                        excess_player.message_queue.put(
                            EndOfConnectionEvent(
                                "Sorry, you are kicked out due to "
                                "a recent rule change."
                            )
                        )
                        excess_player.connected = False
                    self.broadcast(self.get_game_state_event())

    def get_player(
        self,
        player_name: str,
        allow_creation: bool = False,
    ) -> Player:
        with self.lock:

            for player in self.__players:
                if player.name == player_name:
                    return player

            else:

                exception: ApiException | None = None
                if not allow_creation:
                    exception = PlayerNotFoundException(player_name)
                elif self.__started:
                    exception = NewPlayerToStartedGameException(player_name)
                if exception:
                    self.__logger.warn(exception.message)
                    raise exception

                new_player = Player(player_name)
                self.__players.append(new_player)
                self.broadcast(self.get_game_state_event())

                return new_player

    def kick_out_player(
        self,
        *,
        target_name: str,
        operator_name: str | None,
        operator_is_player: bool,
    ) -> None:

        with self.lock:

            target_player = self.get_player(target_name)
            target_player.connected = False
            target_player.message_queue.put(
                EndOfConnectionEvent(
                    format_optional_operator(
                        "Sorry, you are kicked out",
                        operator_name,
                        is_player=operator_is_player,
                    )
                )
            )
            self.__players.remove(target_player)

            if self.started:
                self.__discard_pile.extend(target_player.cards)
                target_player.cards.clear()

        self.__logger.info(
            format_optional_operator(
                f"player#{target_name} is kicked out",
                operator_name,
                is_player=operator_is_player,
            )
        )

    def set_lead_card_info(
        self,
        lead_card: Card,
        *,
        lead_color: BasicCardColor | None = None,
    ) -> None:
        with self.lock:
            if lead_card["type"] == "wild":
                if lead_color is None:
                    raise InvalidLeadCardInfoException(lead_card, lead_color)
                self.__lead_color = lead_color
            else:
                if lead_color is not None:
                    raise InvalidLeadCardInfoException(lead_card, lead_color)
                self.__lead_color = lead_card["color"]
            self.__lead_card = lead_card

    def draw_cards(
        self,
        count: int,
        *,
        player: Player | None,
        allow_shuffle: bool = False,
    ) -> Deck | None:

        drawn_cards: Deck = []

        with self.lock:

            for _ in range(count):

                if not len(self.__draw_pile):

                    if allow_shuffle:
                        self.__draw_pile = self.__discard_pile
                        self.__discard_pile = []
                        self.__logger.debug("Shuffled piles for card drawing.")
                        shuffle(self.__draw_pile)

                    if not len(self.__draw_pile):
                        self.__draw_pile.extend(reversed(drawn_cards))
                        self.broadcast(
                            NotificationEvent(
                                NotificationEvent.DataType(
                                    title="Not Enough Cards",
                                    message="Not enough cards to draw...",
                                )
                            )
                        )
                        self.stop(
                            operator_name=None,
                            operator_is_player=False,
                            state_check_required=False,
                        )
                        return None

                drawn_card = self.__draw_pile.pop()
                drawn_cards.append(drawn_card)

            if player:
                with player.lock:
                    player.cards.extend(drawn_cards)
                    player.message_queue.put(player.get_cards_event())
                self.__logger.debug(
                    f"Cards drawn by player#{player.name}: {drawn_cards!r}"
                )

            self.broadcast(self.get_game_state_event())

        return drawn_cards

    def start(self, player_name: str) -> None:
        with self.lock:

            if self.__started:
                raise GameAlreadyStartedException()

            if len(self.__players) < MIN_PLAYER_CAPACITY:
                raise NotEnoughPlayersException()

            # -- reset card piles --
            self.__draw_pile = create_deck()
            self.__discard_pile = []
            shuffle(self.__draw_pile)

            # -- shuffle players if needed --
            if self.__rules["shuffle_players"]:
                self.__logger.debug("Shuffled players.")
                shuffle(self.__players)

            # -- dispatch initial cards --
            initial_hand_size = self.__rules["initial_hand_size"]
            for player in self.__players:
                player.cards.clear()
            for player in self.__players:
                if not self.draw_cards(initial_hand_size, player=player):
                    return
                player.message_queue.put(player.get_cards_event())

            # -- set lead card --
            lead_card: Card | None = None
            while (lead_card is None) or (lead_card["type"] != "number"):
                lead_card_drawn = self.draw_cards(1, player=None)
                if not lead_card_drawn:
                    return
                else:
                    assert len(lead_card_drawn) == 1
                    lead_card = lead_card_drawn[0]
                    self.__discard_pile.append(lead_card)
            self.set_lead_card_info(lead_card)

            # -- initialize other states --
            self.__current_player_index = 0
            self.__direction = 1
            self.__draw_counter = 0
            self.__skip_counter = 0
            self.__started = True

            message = f"Game started by player#{player_name}."
            self.__logger.info(message)
            self.broadcast(
                NotificationEvent(
                    NotificationEvent.DataType(
                        title="Started!",
                        message=message,
                    )
                )
            )

            self.broadcast(self.get_game_state_event())

    def play(
        self,
        player_name: str,
        card_ids: Sequence[str],
        play_color: BasicCardColor | None,
    ) -> None:
        with self.lock:

            player = self.get_player(player_name)
            expected_player = self.__players[self.__current_player_index]
            if player != expected_player:
                raise NotCurrentPlayerException(expected_player.name)

            player_cards_backup = player.cards.copy()
            cards_out = player.give_out_cards(card_ids)
            n_cards_out = len(cards_out)

            lead_color = self.__lead_color
            lead_card = self.__lead_card
            rules = self.__rules

            if n_cards_out > 0:

                try:
                    if not (lead_color and lead_card):
                        raise InvalidLeadCardInfoException(lead_card, lead_color)
                    check_play(
                        cards_out,
                        play_color,
                        lead_color=lead_color,
                        lead_card=lead_card,
                        skip_counter=self.__skip_counter,
                        rules=rules,
                    )
                    self.set_lead_card_info(cards_out[-1], lead_color=play_color)
                except:
                    player.cards = player_cards_backup
                    raise
                else:

                    play_color_message = (
                        f" (change color to {play_color})" if play_color else ""
                    )
                    self.__logger.info(
                        f"Player#{player.name} played: {cards_out!r}"
                        + play_color_message
                    )
                    self.broadcast(
                        NotificationEvent(
                            NotificationEvent.DataType(
                                title=f"{player.name}'s Play",
                                message=(
                                    ", ".join(map(format_card, cards_out))
                                    + play_color_message
                                ),
                            )
                        )
                    )

                    assert n_cards_out > 0

                    for card in cards_out:
                        if card["type"] == "number":
                            continue
                        elif card["type"] == "function":
                            if card["effect"] == "+2":
                                self.__draw_counter += 2
                                self.__skip_counter += 1
                            elif card["effect"] == "skip":
                                self.__skip_counter += 1
                            elif card["effect"] == "reverse":
                                self.__direction = -self.__direction
                            else:
                                assert_never(card["effect"])
                        elif card["type"] == "wild":
                            if card["effect"] == "+4":
                                self.__draw_counter += 4
                                self.__skip_counter += 1
                            elif card["effect"] == "color":
                                pass
                            else:
                                assert_never(card["effect"])
                        else:
                            assert_never(card["type"])

                    self.__discard_pile.extend(cards_out)

            else:  # n_cards_out == 0
                self.__logger.info(f"Player#{player.name} passed.")
                self.draw_cards(1, player=player, allow_shuffle=True)
                self.broadcast(
                    NotificationEvent(
                        NotificationEvent.DataType(
                            title=f"{player.name}'s Play",
                            message="Pass",
                        )
                    )
                )

            if len(player.cards) == 0:
                if (
                    (not rules["any_last_play"])
                    and (n_cards_out == 1)
                    and (cards_out[0]["type"] != "number")
                ):
                    self.draw_cards(1, player=player, allow_shuffle=True)
                else:
                    self.broadcast(
                        NotificationEvent(
                            NotificationEvent.DataType(
                                title="Game Ended",
                                message=f"Player#{player.name} won!",
                            )
                        )
                    )
                    return self.stop(operator_name=None, operator_is_player=False)

            player_count = len(self.__players)

            if self.__draw_counter > 0:
                next_player_index = (
                    self.__current_player_index + self.__direction
                ) % player_count
                next_player = self.__players[next_player_index]
                self.draw_cards(
                    self.__draw_counter,
                    player=next_player,
                    allow_shuffle=True,
                )
                self.__draw_counter = 0

            self.__current_player_index = (
                self.__current_player_index
                + self.__direction * (self.__skip_counter + 1)
            ) % player_count
            self.__skip_counter = 0

            player.message_queue.put(player.get_cards_event())
            self.broadcast(self.get_game_state_event())

    def stop(
        self,
        operator_name: str | None,
        operator_is_player: bool,
        state_check_required: bool = True,
    ) -> None:
        with self.lock:

            if state_check_required:
                if not self.__started:
                    raise GameNotStartedException()

            self.__started = False
            self.__lead_color = None
            self.__lead_card = None

            for player in self.__players:
                player.last_result = len(player.cards)

            message = format_optional_operator(
                "Game stopped",
                operator_name,
                is_player=operator_is_player,
            )
            self.__logger.info(message)
            self.broadcast(
                NotificationEvent(
                    NotificationEvent.DataType(
                        title="Stopped!",
                        message=message,
                    )
                )
            )

            self.broadcast(self.get_game_state_event())

    def watcher_loop(self) -> None:

        def on_slow(continuous_slow_count: int) -> None:
            if continuous_slow_count >= GAME_WATCHER_SKIP_THRESHOLD:
                self.__logger.warn(
                    "The watcher loop seems to be slow. "
                    f"(continuous_slow_count: {continuous_slow_count})"
                )

        for _ in loop(GAME_WATCHER_INTERVAL, allow_skip=True, on_slow=on_slow):
            with self.lock:

                now = monotonic()

                player_timeout_seconds = PLAYER_TIMEOUT.total_seconds()
                state_changed = False

                for player in self.__players:
                    with player.lock:

                        previous_connected_status = player.connected

                        player.connected = False
                        if player.subscription_token:
                            if player.last_pending_timestamp:
                                pending_time = now - player.last_pending_timestamp
                                if pending_time < player_timeout_seconds:
                                    player.connected = True
                            else:
                                player.connected = True

                        if player.connected != previous_connected_status:
                            state_changed = True
                            if not player.connected:
                                self.__logger.info(
                                    f"Disconnected from player#{player.name}. "
                                    f"(subscription_token: {player.subscription_token})"
                                )
                                player.subscription_token = ""
                                if not self.__started:
                                    self.__players.remove(player)

                if state_changed:
                    self.broadcast(self.get_game_state_event())


game = Game()



"""UNO right in your terminal, with both server and client."""

import click

from tuno.client import start_client
from tuno.server import start_server

__version__ = "0.1.0"


@click.group(
    help=__doc__,
    context_settings=dict(
        help_option_names=["-h", "--help"],
    ),
)
@click.version_option(
    __version__,
    "-v",
    "--version",
)
def tuno() -> None:
    pass


tuno.add_command(start_client)
tuno.add_command(start_server)



from abc import abstractmethod
from inspect import get_annotations
from typing import (
    Annotated,
    NamedTuple,
    Protocol,
    TypedDict,
    get_origin,
    runtime_checkable,
)

from tuno.server.exceptions import ApiException
from tuno.shared.constraints import (
    MAX_INITIAL_HAND_SIZE,
    MAX_PLAYER_CAPACITY,
    MIN_INITIAL_HAND_SIZE,
    MIN_PLAYER_CAPACITY,
)


@runtime_checkable
class RuleValidator(Protocol):

    rule_name: str

    def __init__(self, rule_name: str) -> None:
        super().__init__()
        self.rule_name = rule_name

    @abstractmethod
    def validate(self, value: object) -> None: ...


class RuleValidationException(ApiException):

    def __init__(self, message: str) -> None:
        super().__init__(400, message)


class IntRangeRuleValidator(RuleValidator):

    min: float
    max: float

    def __init__(self, rule_name: str, min: float, max: float) -> None:
        super().__init__(rule_name)
        self.min = min
        self.max = max

    def validate(self, value: object) -> None:
        if not isinstance(value, int):
            raise RuleValidationException(
                f"{self.rule_name} must be an integer, " f"got: {type(value).__name__}"
            )
        if not self.min <= value <= self.max:
            raise RuleValidationException(
                f"{self.rule_name} must be in range "
                f"[{self.min}, {self.max}], got: {value}"
            )


class GameRules(TypedDict):

    player_capacity: Annotated[
        int,
        f"{MIN_PLAYER_CAPACITY}~{MAX_PLAYER_CAPACITY}",
        IntRangeRuleValidator(
            "player_capacity",
            MIN_PLAYER_CAPACITY,
            MAX_PLAYER_CAPACITY,
        ),
    ]

    shuffle_players: Annotated[
        bool,
        "shuffle players before starting",
        None,
    ]

    initial_hand_size: Annotated[
        int,
        f"{MIN_INITIAL_HAND_SIZE}~{MAX_INITIAL_HAND_SIZE}",
        IntRangeRuleValidator(
            "initial_hand_size",
            MIN_INITIAL_HAND_SIZE,
            MAX_INITIAL_HAND_SIZE,
        ),
    ]

    any_last_play: Annotated[
        bool,
        "allow non-number card as last play",
        None,
    ]


class RuleAnnotation(NamedTuple):
    type: type
    hint: str
    validator: RuleValidator | None


def get_rule_annotation(key: str) -> RuleAnnotation:

    annotations = get_annotations(GameRules)

    if key not in annotations:
        raise RuleValidationException(f"Unknown rule: {key}")

    type_annotation = annotations[key]
    if get_origin(type_annotation) is not Annotated:
        raise RuleValidationException(f"Invalid rule annotation: {type_annotation}")

    type_metadata: tuple[object, object, object] = type_annotation.__metadata__
    if (
        (len(type_metadata) != 2)
        or (not isinstance(type_metadata[0], str))
        or not (
            (type_metadata[1] is None) or isinstance(type_metadata[1], RuleValidator)
        )
    ):
        raise RuleValidationException(
            f"Invalid metadata for rule `{key}`: {type_metadata!r}"
        )

    return RuleAnnotation(
        type=type_annotation.__origin__,
        hint=type_metadata[0],
        validator=type_metadata[1],
    )


def check_rule_update(key: str, value: object) -> None:

    rule_annotation = get_rule_annotation(key)

    if not isinstance(value, rule_annotation.type):
        raise RuleValidationException(
            f"Invalid type for rule `{key}`: "
            f"expected {rule_annotation.type}, got {type(value)}"
        )

    if rule_annotation.validator:
        rule_annotation.validator.validate(value)



from typing import TYPE_CHECKING, cast

from flask import Blueprint, request
from flask.typing import ResponseReturnValue

from tuno.server.exceptions import InvalidRequestBodyException
from tuno.server.utils.checkers import check_player_name
from tuno.server.utils.Logger import Logger
from tuno.shared.deck import BasicCardColor, basic_card_colors

logger = Logger(__name__)


def setup(blueprint: Blueprint) -> None:

    @blueprint.post("/<player_name>/play")
    def play(player_name: str) -> ResponseReturnValue:

        check_player_name(player_name)

        card_ids = request.json

        if not (
            isinstance(card_ids, list)
            and all(isinstance(card_id, str) for card_id in card_ids)
        ):
            raise InvalidRequestBodyException(
                f"expected a list of strings as card_ids, got {card_ids!r}."
            )

        play_color = request.args.get("color", None)
        if play_color is not None and play_color not in basic_card_colors:
            raise InvalidRequestBodyException(
                f"expected a valid card color, got {play_color!r}."
            )
        if TYPE_CHECKING:
            play_color = cast(BasicCardColor, play_color)

        from tuno.server.models.Game import game

        game.play(player_name, card_ids, play_color)

        return ("Play accepted.", 200)



from tuno.server.exceptions import InvalidPlayerNameException
from tuno.shared.constraints import PLAYER_NAME_PATTERN


def check_player_name(player_name: str) -> None:
    if not PLAYER_NAME_PATTERN.fullmatch(player_name):
        raise InvalidPlayerNameException(player_name)



from dataclasses import dataclass
from functools import partialmethod

from textual import on
from textual.app import App
from textual.message import Message
from textual.screen import ModalScreen

from tuno.client.config import (
    NOTIFICATION_TIMEOUT_DEFAULT,
    NOTIFICATION_TIMEOUT_ERROR,
)
from tuno.shared.deck import Deck
from tuno.shared.sse_events import GameStateEvent

from .components.ConnectScreen import ConnectScreen
from .components.InGameScreen import InGameScreen
from .components.PendingScreen import PendingScreen
from .UnoClient import UnoClient


class UnoApp(App[object]):

    NOTIFICATION_TIMEOUT = NOTIFICATION_TIMEOUT_DEFAULT.total_seconds()
    MODES = {
        "connect": ConnectScreen,
        "pending": PendingScreen,
        "in-game": InGameScreen,
    }

    @dataclass
    class GameStateUpdate(Message):
        game_state: GameStateEvent.DataType

    @dataclass
    class CardsUpdate(Message):
        cards: Deck

    client: UnoClient | None

    notify_error = partialmethod(
        App.notify,
        severity="error",
        timeout=NOTIFICATION_TIMEOUT_ERROR.total_seconds(),
    )

    def on_mount(self) -> None:
        self.client = UnoClient(self)
        self.switch_mode("connect")

    def on_unmount(self) -> None:
        client = self.client
        assert client is not None
        with client.subscription_lock:
            if client.subscription:
                client.subscription.close()
                client.subscription = None
                self.log.debug("Subscription detached.")

    @on(GameStateUpdate)
    async def on_game_state_update(self, message: GameStateUpdate) -> None:

        game_state = message.game_state

        if game_state:
            if game_state["started"]:
                await self.switch_mode("in-game")
            else:
                await self.switch_mode("pending")
            for modal_screen in self.query(ModalScreen):
                modal_screen.dismiss()
        else:
            await self.switch_mode("connect")

        if hasattr(self.screen, "game_state"):
            setattr(self.screen, "game_state", game_state)
            self.log.debug("Updated game state on current screen.")

    @on(CardsUpdate)
    def on_cards_update(self, message: CardsUpdate) -> None:
        screen = self.screen
        if isinstance(screen, InGameScreen):
            screen.cards = message.cards
            self.log.debug("Updated game state on InGameScreen.")



from datetime import timedelta
from typing import Final

# -- Connection Config --
SSE_MAX_RETRIES: Final = 2
SSE_TIMEOUT = timedelta(seconds=10)

# -- Notification Config --
NOTIFICATION_TIMEOUT_DEFAULT = timedelta(seconds=3)
NOTIFICATION_TIMEOUT_ERROR = timedelta(seconds=10)

# -- Environment Config --
ENV_KEY_CONNECTION: Final = "UNO_CONNECTION"
ENV_KEY_PLAYER_NAME: Final = "UNO_PLAYER_NAME"
ENV_KEY_SERVER_ADDRESS: Final = "UNO_SERVER_ADDRESS"



from dataclasses import dataclass
from typing import Final, cast

from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal, Right, Vertical, VerticalScroll
from textual.events import DescendantBlur, DescendantFocus
from textual.message import Message
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.validation import Function as FunctionValidator
from textual.widget import Widget
from textual.widgets import Button, Footer, Header, Input, Label, Switch

from tuno.client.utils.LoadingContext import LoadingContext
from tuno.shared.rules import (
    GameRules,
    RuleValidationException,
    get_rule_annotation,
)


class RulesFormItem(Widget):

    __CLASS_FOCUS: Final = "focus"
    __CLASS_INVALID: Final = "invalid"

    readonly: bool
    rule_key: str
    rule_type: type
    rule_hint: str
    textual_validator: FunctionValidator | None
    focused_descendants: set[Widget]

    def __init__(self, rule_key: str, *, readonly: bool) -> None:
        super().__init__()

        self.readonly = readonly
        self.rule_key = rule_key
        self.focused_descendants = set()

        rule_annotation = get_rule_annotation(rule_key)
        self.rule_type = rule_annotation.type
        self.rule_hint = rule_annotation.hint

        rule_validator = rule_annotation.validator
        if rule_validator:

            def validator_wrapper(value: str | bool) -> bool:
                if isinstance(value, str):
                    if value != "":
                        value = rule_annotation.type(value)
                    else:
                        value = rule_annotation.type()
                try:
                    assert rule_validator is not None
                    rule_validator.validate(value)
                except RuleValidationException as exception:
                    textual_validator.failure_description = exception.message
                    return False
                else:
                    return True

            textual_validator = FunctionValidator(validator_wrapper)
            self.textual_validator = textual_validator

        else:
            self.textual_validator = None

    def compose(self) -> ComposeResult:

        from tuno.client.UnoApp import UnoApp

        app = self.app
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None
        assert client.game_state is not None

        rule_key = self.rule_key
        rule_hint = self.rule_hint
        rule_text = f"{rule_key} ({rule_hint})"
        yield Label(rule_text)

        self.tooltip = rule_text

        rule_type = self.rule_type
        current_rules = client.game_state["rules"]
        current_value = current_rules[rule_key]  # type: ignore[literal-required]
        input_widget: Widget
        if rule_type is int:
            input_widget = Input(
                value=str(current_value),
                type="integer",
                classes="rule-input",
                validators=self.textual_validator,
                placeholder=rule_hint,
                tooltip=rule_hint,
            )
        elif rule_type is bool:
            input_widget = Switch(
                value=current_value,
                classes="rule-input",
                tooltip=rule_hint,
            )
        else:
            raise NotImplementedError()
        input_widget.disabled = self.readonly
        yield Right(input_widget)

    @on(DescendantFocus)
    def on_descendant_focus(self, event: DescendantFocus) -> None:
        self.focused_descendants.add(event.control)
        self.set_class(len(self.focused_descendants) > 0, self.__CLASS_FOCUS)

    @on(DescendantBlur)
    def on_descendant_blur(self, event: DescendantBlur) -> None:
        self.focused_descendants.discard(event.control)
        self.set_class(len(self.focused_descendants) > 0, self.__CLASS_FOCUS)

    @on(Input.Changed)
    def on_input_changed(self, event: Input.Changed) -> None:
        event.control.set_class(
            not event.control.is_valid,
            self.__CLASS_INVALID,
        )

    @on(DescendantFocus)
    def get_rule_value(self) -> object:
        input_widget = self.query_exactly_one(".rule-input")
        if isinstance(input_widget, Input):
            if input_widget.is_valid:
                return self.rule_type(input_widget.value)
            else:
                return None
        elif isinstance(input_widget, Switch):
            return input_widget.value
        else:
            return None

    def set_rule_value(self, value: object) -> None:
        input_widget = self.query_exactly_one(".rule-input")
        if isinstance(input_widget, Input):
            assert isinstance(value, self.rule_type)
            input_widget.value = str(value)
        elif isinstance(input_widget, Switch):
            assert isinstance(value, bool)
            input_widget.value = value
        else:
            raise NotImplementedError()


class RulesForm(VerticalScroll):

    @dataclass
    class FormValidationChanged(Message):
        all_valid: bool

    invalid_inputs: reactive[set[Input]] = reactive(set, init=False)
    all_valid: reactive[bool] = reactive(True, init=False)

    readonly: bool

    def __init__(self, *, readonly: bool) -> None:
        super().__init__()
        self.readonly = readonly

    def compute_all_valid(self) -> bool:
        return len(self.invalid_inputs) == 0

    def compose(self) -> ComposeResult:
        for key in GameRules.__annotations__.keys():
            yield RulesFormItem(key, readonly=self.readonly)

    def on_mount(self) -> None:
        invalid_input_detected = False
        for input_widget in self.query(Input):
            if not input_widget.is_valid:
                invalid_input_detected = True
                self.invalid_inputs.add(input_widget)
        if invalid_input_detected:
            self.mutate_reactive(RulesForm.invalid_inputs)

    @on(Input.Changed)
    def on_input_changed(self, message: Input.Changed) -> None:
        if message.control.is_valid:
            self.invalid_inputs.discard(message.control)
        else:
            self.invalid_inputs.add(message.control)
        self.mutate_reactive(RulesForm.invalid_inputs)

    def watch_all_valid(self, all_valid: bool) -> None:
        self.post_message(self.FormValidationChanged(all_valid))


class RulesScreen(ModalScreen[object]):

    TITLE = "UNO"
    CSS_PATH = "styles.tcss"
    BINDINGS = [
        ("ctrl+s", "screen.submit", "Submit"),
        ("escape", "screen.dismiss", "Cancel"),
    ]

    readonly: bool

    def __init__(self, *, readonly: bool) -> None:
        super().__init__()
        self.readonly = readonly

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        yield Vertical(
            Label("Game Rules", id="rules-title"),
            RulesForm(readonly=self.readonly),
            Horizontal(
                Button(
                    "Submit",
                    id="rules-submit",
                    variant="primary",
                    action="screen.submit",
                    disabled=self.readonly,
                ),
                Button(
                    "Reset",
                    variant="error",
                    action="screen.reset",
                    disabled=self.readonly,
                ),
                Button("Cancel", action="screen.dismiss"),
                id="rules-actions",
            ),
            id="rules-window",
        )
        yield Footer()

    def on_mount(self) -> None:

        from tuno.client.UnoApp import UnoApp

        app = cast(UnoApp, self.app)
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None

        self.sub_title = client.get_connection_display()

    @on(RulesForm.FormValidationChanged)
    def on_form_validation_changed(
        self,
        message: RulesForm.FormValidationChanged,
    ) -> None:
        self.query_exactly_one("#rules-submit", Button).disabled = not message.all_valid

    @work(thread=True)
    def action_submit(self) -> None:

        if self.readonly:
            return

        from tuno.client.UnoApp import UnoApp

        app = self.app
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None
        assert client.game_state is not None

        current_rules = client.game_state["rules"]
        modified_rules: dict[str, object] = {}
        for form_item in self.query(RulesFormItem):
            rule_key = form_item.rule_key
            old_value = current_rules[rule_key]  # type: ignore[literal-required]
            new_value = form_item.get_rule_value()
            if new_value != old_value:
                modified_rules[rule_key] = new_value

        if len(modified_rules) == 0:
            self.app.notify(
                "Rules are unchanged.",
                title="Submission Ignored",
                severity="warning",
            )
            return

        with LoadingContext("Updating rules...", app=app):
            client.update_rules(modified_rules)

    def action_reset(self) -> None:

        from tuno.client.UnoApp import UnoApp

        app = self.app
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None
        assert client.game_state is not None

        current_rules = client.game_state["rules"]
        for form_item in self.query(RulesFormItem):
            rule_key = form_item.rule_key
            rule_value = current_rules[rule_key]  # type: ignore[literal-required]
            form_item.set_rule_value(rule_value)



from textual.reactive import reactive
from textual.widgets import Label

from tuno.shared.deck import Card


class CardLabel(Label):

    DEFAULT_CSS = """
    CardLabel {
        width: 100%;
        height: 100%;
        content-align-horizontal: center;
        color: $text;
    }
    """

    data: reactive[Card | None] = reactive(None)

    def watch_data(self, data: Card | None) -> None:

        if data is None:
            self.styles.background = "gray"
            self.update("???")
            self.tooltip = None
            return

        self.styles.background = f"{data['color']} 90%"

        content = ""
        if data["type"] == "number":
            content += str(data["number"])
        else:
            content += data["effect"]
        self.update(content)
        self.tooltip = content



from typing import cast

from textual import work
from textual.app import ComposeResult
from textual.containers import HorizontalScroll
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import Button, Footer, Header

from tuno.client.components.Players import Players
from tuno.client.components.RulesScreen import RulesScreen
from tuno.client.utils.LoadingContext import LoadingContext
from tuno.shared.sse_events import GameStateEvent


class PendingScreen(Screen[object]):

    TITLE = "UNO"
    CSS_PATH = "styles.tcss"
    BINDINGS = [
        ("ctrl+g", "start_game", "Start"),
        ("ctrl+r", "show_rules", "Rules"),
    ]

    game_state: reactive[GameStateEvent.DataType | None] = reactive(None)

    def compose(self) -> ComposeResult:

        actions_container = HorizontalScroll(
            Button(
                "Start",
                id="action-start-game",
                variant="success",
                action="screen.start_game",
            ),
            Button(
                "Rules",
                id="action-show-rules",
                action="screen.show_rules",
            ),
            id="actions",
        )
        actions_container.border_title = "Actions"

        yield Header(show_clock=True)
        yield Players().data_bind(PendingScreen.game_state)
        yield actions_container
        yield Footer()

    def watch_game_state(self, game_state: GameStateEvent.DataType | None) -> None:

        from tuno.client.UnoApp import UnoApp

        app = cast(UnoApp, self.app)
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None
        self.sub_title = client.get_connection_display()

    def action_show_rules(self) -> None:
        self.app.push_screen(RulesScreen(readonly=False))

    @work(thread=True)
    def action_start_game(self) -> None:

        from tuno.client.UnoApp import UnoApp

        app = cast(UnoApp, self.app)
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None

        with LoadingContext("Starting game...", app=app):
            client.start_game()



from collections.abc import Generator, Sequence
from contextlib import contextmanager
from queue import Queue
from threading import RLock
from time import monotonic

from tuno.server.config import PLAYER_MESSAGE_QUEUE_SIZE
from tuno.server.exceptions import CardIdsNotFoundException
from tuno.server.utils.Logger import Logger
from tuno.shared.deck import Deck
from tuno.shared.sse_events import CardsEvent, ServerSentEvent


class Player:

    name: str
    cards: Deck
    last_result: int
    message_queue: Queue[ServerSentEvent]
    lock: RLock
    connected: bool  # set by game watcher
    subscription_token: str
    __logger: Logger
    __last_pending_timestamp: float | None
    __last_sent_timestamp: float | None

    def __init__(self, name: str) -> None:

        self.name = name
        self.cards = []
        self.last_result = -1
        self.message_queue = Queue(PLAYER_MESSAGE_QUEUE_SIZE)
        self.lock = RLock()
        self.connected = False
        self.subscription_token = ""
        self.__logger = Logger(f"{__name__}#{name}")
        self.__last_pending_timestamp = None
        self.__last_sent_timestamp = None

        self.__logger.debug(f"player#{name} created")

    @property
    def last_pending_timestamp(self) -> float | None:
        return self.__last_pending_timestamp

    @property
    def last_sent_timestamp(self) -> float | None:
        return self.__last_sent_timestamp

    @contextmanager
    def message_context(self) -> Generator[None]:
        with self.lock:
            self.__last_pending_timestamp = monotonic()
            yield
            self.__last_pending_timestamp = None
            self.__last_sent_timestamp = monotonic()

    def get_cards_event(self) -> CardsEvent:
        return CardsEvent(self.cards)

    def give_out_cards(self, card_ids: Sequence[str]) -> Deck:

        if len(card_ids) == 0:
            return []

        cards_ids_remaining = set(card_ids)
        cards_left: Deck = []
        cards_out: Deck = []

        with self.lock:

            for card in self.cards:
                if card["id"] in cards_ids_remaining:
                    cards_ids_remaining.remove(card["id"])
                    cards_out.append(card)
                else:
                    cards_left.append(card)

            if len(cards_ids_remaining) > 0:
                raise CardIdsNotFoundException(cards_ids_remaining)

            self.cards = cards_left

        return cards_out



from collections.abc import Generator
from queue import Empty
from secrets import token_hex
from time import monotonic

from flask import Blueprint, Response
from flask.typing import ResponseReturnValue

from tuno.server.config import (
    HEARTBEAT_GAP,
    MAX_MESSAGES_PER_LOOP,
    SUBSCRIPTION_LOOP_INTERVAL,
    SUBSCRIPTION_LOOP_SLOW_THRESHOLD,
    SUBSCRIPTION_TOKEN_BYTES,
)
from tuno.server.utils.checkers import check_player_name
from tuno.server.utils.Logger import Logger
from tuno.shared.loop import loop
from tuno.shared.sse_events import (
    EndOfConnectionEvent,
    GameStateEvent,
    SubscriptionChangeEvent,
)

logger = Logger(__name__)


def setup(blueprint: Blueprint) -> None:

    HEARTBEAT_GAP_SECONDS = HEARTBEAT_GAP.total_seconds()

    @blueprint.get("/<player_name>")
    def player_subscription(player_name: str) -> ResponseReturnValue:

        check_player_name(player_name)

        from tuno.server.models.Game import game

        player = game.get_player(player_name, allow_creation=True)
        subscription_token = token_hex(SUBSCRIPTION_TOKEN_BYTES)
        player.subscription_token = subscription_token

        def event_generator() -> Generator[str]:

            # send first heartbeat
            with player.message_context():
                yield f":\n\n"
            logger.info(
                f"Connected with player#{player_name}. "
                f"(subscription_token: {subscription_token})"
            )

            # send initial states
            with player.message_context():
                yield game.get_game_state_event().to_sse()
                yield player.get_cards_event().to_sse()
            logger.debug(
                f"Sent initial states to player#{player_name}. "
                f"(subscription_token: {subscription_token})"
            )

            # slow iteration detection
            def on_slow(continuous_slow_count: int) -> None:
                if continuous_slow_count >= SUBSCRIPTION_LOOP_SLOW_THRESHOLD:
                    logger.warn(
                        "The subscription loop seems to be slow. "
                        f"(subscription_token: {subscription_token}; "
                        f"continuous_slow_count: {continuous_slow_count})"
                    )

            # message loop
            for _ in loop(SUBSCRIPTION_LOOP_INTERVAL, on_slow=on_slow):
                with player.lock:

                    for _ in range(MAX_MESSAGES_PER_LOOP):

                        if player.subscription_token != subscription_token:
                            break

                        try:
                            event = player.message_queue.get_nowait()
                        except Empty:
                            break
                        else:

                            with player.message_context():
                                yield event.to_sse()
                            player.message_queue.task_done()
                            logger.debug(
                                f"Event sent to player#{player_name} "
                                f"(subscription_token: {subscription_token}): "
                                + repr(event)
                            )

                            if isinstance(event, EndOfConnectionEvent):
                                player.subscription_token = ""
                                logger.debug(
                                    "Stopped subscription from "
                                    f"player#{player_name} due to the presence"
                                    f"of an {EndOfConnectionEvent.__name__}. "
                                    f"(subscription_token: {subscription_token})"
                                )
                                return

                    if player.subscription_token != subscription_token:
                        break

                    if (
                        monotonic() - (player.last_sent_timestamp or 0)
                        >= HEARTBEAT_GAP_SECONDS
                    ):
                        with player.message_context():
                            yield f":\n\n"  # heartbeat
                        # logger.debug(f"Heartbeat sent to player#{player_name}.")

            if player.subscription_token:  # replaced by another subscription
                event = SubscriptionChangeEvent()
                with player.message_context():
                    yield event.to_sse()
                logger.debug(
                    f"Event sent to player#{player_name} "
                    f"(subscription_token: {subscription_token}): " + repr(event)
                )

            logger.info(
                f"Subscription stopped for player#{player_name}. "
                f"(subscription_token: {subscription_token})"
            )

        response = Response(event_generator(), mimetype="text/event-stream")

        @response.call_on_close
        def on_close() -> None:
            with player.lock:
                if player.subscription_token == subscription_token:
                    player.subscription_token = ""
                    logger.info(
                        f"Disconnected from player#{player_name}. "
                        f"(subscription_token: {subscription_token})"
                    )

        return response



from tuno.server.exceptions import ApiException
from tuno.shared.deck import BasicCardColor, Card, Deck
from tuno.shared.rules import GameRules


class InvalidPlayException(ApiException):

    def __init__(self, message: str) -> None:
        super().__init__(400, message)


def check_play(
    play: Deck,
    play_color: BasicCardColor | None,
    *,
    lead_color: BasicCardColor,
    lead_card: Card,
    skip_counter: int,
    rules: GameRules,
) -> None:

    if len(play) == 0:
        raise InvalidPlayException("At least one card must be given in a play.")

    if len(play) > 1:
        raise InvalidPlayException("Only one card can be played at a time.")

    if skip_counter > 0:
        raise InvalidPlayException("Skipped players cannot play.")

    played_card = play[0]

    if played_card["type"] != "wild":

        if played_card["color"] == lead_color:
            return

        if (
            (lead_card["type"] == "number")
            and (played_card["type"] == "number")
            and (played_card["number"] == lead_card["number"])
        ):
            return

        if (
            (lead_card["type"] == "function")
            and (played_card["type"] == "function")
            and (played_card["effect"] == lead_card["effect"])
        ):
            return

        raise InvalidPlayException(
            "Non-wild-card play must match the lead color, number or effect."
        )

    else:

        if not play_color:
            raise InvalidPlayException(
                "Wild cards must be played with a color claimed."
            )

        return

    raise NotImplementedError()



from typing import TYPE_CHECKING, Final, cast

from textual.app import ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Label

from tuno.shared.sse_events import GameStateEvent


class PlayerCard(Widget):

    __CLASS_PLAYER_ACTIVE: Final = "active"
    __CLASS_PLAYER_CONNECTED: Final = "connected"
    __CLASS_PLAYER_DISCONNECTED: Final = "disconnected"

    DEFAULT_CSS = """
    PlayerCard {
        height: 4;
        border: round $panel;
        padding: 0 1;
        layout: grid;
        grid-size: 2 2;
        grid-rows: 1 1;
        grid-columns: 2 1fr;

        &.active {
            border: heavy $success;
        }

        .player-card-connected {
            width: 1;
            color: $foreground-muted;

            &.connected {
                color: $text-success;
            }
            &.disconnected {
                color: $text-error;
            }
        }

        .player-card-name {
            width: 100%;
        }

        .player-card-count {
            column-span: 2;
            align-horizontal: right;

            .player-card-count-label {
                color: $foreground-muted;
            }
        }
    }
    """

    active: reactive[bool] = reactive(False)
    data: reactive[GameStateEvent.PlayerDataType | None] = reactive(None)

    def compose(self) -> ComposeResult:

        label_connected = Label("●", classes="player-card-connected")
        label_connected.tooltip = "Connection: N/A"

        label_name = Label("???", classes="player-card-name")
        label_name.tooltip = "Player name"

        yield label_connected
        yield label_name
        yield Horizontal(
            Label(
                "Cards: ",
                classes="player-card-count-label",
            ),
            Label(
                "...",
                classes="player-card-count-value",
            ),
            classes="player-card-count",
        )

    def watch_active(self, active: bool) -> None:
        if active:
            self.add_class(self.__CLASS_PLAYER_ACTIVE)
        else:
            self.remove_class(self.__CLASS_PLAYER_ACTIVE)

    def watch_data(self, data: GameStateEvent.PlayerDataType | None) -> None:

        label_name = self.query_exactly_one(
            ".player-card-name",
            Label,
        )
        if data:
            label_name.update(data["name"])
            label_name.tooltip = f"Player name: {data["name"]}"
            if data["connected"]:
                label_name.add_class(self.__CLASS_PLAYER_CONNECTED)
                label_name.remove_class(self.__CLASS_PLAYER_DISCONNECTED)
            else:
                label_name.remove_class(self.__CLASS_PLAYER_CONNECTED)
                label_name.add_class(self.__CLASS_PLAYER_DISCONNECTED)
        else:
            label_name.update("???")
            label_name.remove_class(self.__CLASS_PLAYER_CONNECTED)
            label_name.add_class(self.__CLASS_PLAYER_DISCONNECTED)
            label_name.tooltip = "Player name"

        label_connected = self.query_exactly_one(
            ".player-card-connected",
            Label,
        )
        if data:
            if data["connected"]:
                label_connected.tooltip = "Connected"
                label_connected.add_class(self.__CLASS_PLAYER_CONNECTED)
                label_connected.remove_class(self.__CLASS_PLAYER_DISCONNECTED)
            else:
                label_connected.tooltip = "Disconnected"
                label_connected.remove_class(self.__CLASS_PLAYER_CONNECTED)
                label_connected.add_class(self.__CLASS_PLAYER_DISCONNECTED)
        else:
            label_connected.tooltip = "Connection: N/A"
            label_connected.remove_class(
                self.__CLASS_PLAYER_CONNECTED,
                self.__CLASS_PLAYER_DISCONNECTED,
            )

        label_card_count_value = self.query_exactly_one(
            ".player-card-count-value",
            Label,
        )
        card_count_display = "N/A"
        if data:
            card_count = data["card_count"]
            if card_count >= 0:
                card_count_display = str(card_count)
        label_card_count_value.update(card_count_display)


class Players(VerticalScroll):

    BORDER_TITLE = "Players"
    DEFAULT_CSS = """
    Players {
        padding: 1 2;
        layout: grid;
        grid-size: 3;
        grid-rows: 4;
        grid-gutter: 0 1;

        LoadingIndicator {
            background: transparent;
        }
    }
    """

    game_state: reactive[GameStateEvent.DataType | None] = reactive(None)

    async def watch_game_state(
        self,
        game_state: GameStateEvent.DataType | None,
    ) -> None:

        if not game_state:
            self.remove_children()
            self.loading = True
            return

        self.loading = False

        async with self.batch():

            children = self.children
            assert all(isinstance(child, PlayerCard) for child in children)
            if TYPE_CHECKING:
                children = cast(list[PlayerCard], children)

            players = game_state["players"]
            current_child_count = len(children)
            for i, player_data in enumerate(players[:current_child_count]):
                children[i].data = player_data
                children[i].active = game_state["started"] and (
                    i == game_state["current_player_index"]
                )

            player_count = len(players)
            if player_count > current_child_count:
                new_players = players[current_child_count:]
                new_children: list[PlayerCard] = []
                for player_data in new_players:
                    new_children.append(PlayerCard())
                await self.mount(*new_children)
                for j, (child, data) in enumerate(
                    zip(new_children, new_players, strict=True)
                ):
                    child.data = data
                    i = current_child_count + j
                    child.active = game_state["started"] and (
                        i == game_state["current_player_index"]
                    )
            elif player_count < current_child_count:
                self.remove_children(children[player_count:])



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tuno.client.UnoApp import UnoApp


def handler(parsed_data: str, app: "UnoApp") -> None:

    client = app.client
    assert client

    client.reset(parsed_data)
    app.switch_mode("connect")

    app.notify_error(
        parsed_data,
        title="Connection Stopped",
    )



import json
from collections.abc import Callable, Mapping
from threading import RLock, Thread
from typing import TYPE_CHECKING

from requests import RequestException, get, post, put
from requests_sse import EventSource
from textual import log

from tuno.client.config import SSE_MAX_RETRIES, SSE_TIMEOUT
from tuno.client.event_handlers import EventHandlerMap, load_event_handler_map
from tuno.client.utils.ApiContext import ApiContext
from tuno.shared.deck import Deck
from tuno.shared.sse_events import GameStateEvent

if TYPE_CHECKING:
    from tuno.client.UnoApp import UnoApp


class UnoClient:

    app: "UnoApp"
    server_address: str
    player_name: str
    game_state: GameStateEvent.DataType | None
    cards: Deck
    subscription: EventSource | None
    subscription_lock: RLock
    subscription_thread: Thread | None
    event_handler_map: EventHandlerMap

    def __init__(self, app: "UnoApp") -> None:
        super().__init__()
        self.app = app
        self.server_address = ""
        self.player_name = ""
        self.game_state = None
        self.cards = []
        self.subscription = None
        self.subscription_lock = RLock()
        self.subscription_thread = None
        self.event_handler_map = load_event_handler_map()

    def get_connection_display(self) -> str:
        with self.subscription_lock:
            if self.subscription is None:
                raise RuntimeError("Server not connected!")
            assert self.player_name
            assert self.server_address
            return f"{self.player_name}@{self.server_address}"

    def reset(self, message: str | None) -> None:

        with self.subscription_lock:
            if self.subscription:
                self.subscription.close()
                self.subscription = None

        self.server_address = ""
        self.player_name = ""
        self.game_state = None
        self.cards = []

        log_message = "Client reset."
        if message:
            log_message += f" ({message})"
        self.app.log.info(log_message)

    def get_api_url(self, api_path: str) -> str:
        assert self.server_address
        return f"http://{self.server_address}/api{api_path}"

    def subscribe(
        self,
        *,
        server_address: str,
        player_name: str,
        on_success: Callable[[], None],
        on_failure: Callable[[str], None],
    ) -> None:

        self.server_address = server_address
        self.player_name = player_name

        url = self.get_api_url(f"/player/{player_name}")
        log.info(f'Connecting to "{url}"...')

        event_source = EventSource(
            url,
            max_connect_retry=SSE_MAX_RETRIES,
            timeout=SSE_TIMEOUT.total_seconds(),
        )

        with self.subscription_lock:
            self.subscription = event_source

        try:

            with event_source:

                log.info(f'Connected to "{server_address}".')
                on_success()

                for event in event_source:

                    if self.subscription != event_source:
                        break

                    log.debug("Received server-sent event:", event)

                    handler = self.event_handler_map.get(event.type or "")
                    if not handler:
                        error_message = f"Unexpected event type: {event.type!r}"
                        log.error(error_message)
                        self.app.notify_error(
                            error_message,
                            title="Unknown Event Type",
                        )
                        continue

                    parsed_data: object = None
                    if event.data:
                        try:
                            parsed_data = json.loads(event.data)
                        except json.JSONDecodeError as error:
                            error_message = f"Invalid JSON data: {event.data!r}"
                            log.error(error_message)
                            log.error(error)
                            self.app.notify_error(
                                error_message,
                                title="Invalid Event Data",
                            )
                            continue

                    handler(parsed_data, self.app)

        except RequestException as error:
            log.error(error)
            if error.response:
                error_message = error.response.text
            else:
                error_message = repr(error)
            on_failure(error_message)

        except AttributeError as error:
            if (error.obj is None) and (error.name == "read"):
                # HACK: After being closed, the SSE client may block until next
                # incoming line and raise an AttributeError for next read.
                # Therefore, this clause suppresses such error so that
                # the app can quit without error reported.
                event_source.close()
                pass
            else:
                raise  # propagate other errors

    def connect(
        self,
        *,
        server_address: str,
        player_name: str,
        on_success: Callable[[], None],
        on_failure: Callable[[str], None],
    ) -> None:
        self.subscription_thread = Thread(
            target=self.subscribe,
            kwargs=dict(
                server_address=server_address,
                player_name=player_name,
                on_success=on_success,
                on_failure=on_failure,
            ),
        )
        self.subscription_thread.start()

    def update_rules(self, modified_rules: Mapping[str, object]) -> None:
        api_url = self.get_api_url("/game/rules")
        assert self.player_name
        with ApiContext("Rule Update Failed", app=self.app):
            response = put(
                api_url,
                json=modified_rules,
                params={
                    "player_name": self.player_name,
                },
            )
            response.raise_for_status()

    def start_game(self) -> None:

        player_name = self.player_name
        assert player_name

        api_url = self.get_api_url(f"/game/start")

        with ApiContext("Start Failed", app=self.app):
            response = put(
                api_url,
                params={
                    "player_name": player_name,
                },
            )
            response.raise_for_status()

    def play(self, card_ids: list[str], color: str | None) -> None:

        player_name = self.player_name
        assert player_name

        api_url = self.get_api_url(f"/player/{player_name}/play")

        with ApiContext("Play Failed", app=self.app):
            response = post(
                api_url,
                params={
                    "color": color,
                },
                json=card_ids,
            )
            response.raise_for_status()

    def stop_game(self) -> None:

        player_name = self.player_name
        assert player_name

        api_url = self.get_api_url(f"/game/stop")

        with ApiContext("Stop Failed", app=self.app):
            response = put(
                api_url,
                params={
                    "player_name": player_name,
                },
            )
            response.raise_for_status()



def format_optional_operator(
    message: str,
    operator_name: str | None,
    *,
    is_player: bool,
    preposition: str = "by",
) -> str:
    if is_player:
        return f"{message} {preposition} player#{operator_name}."
    elif operator_name:
        return f"{message} {preposition} {operator_name}."
    else:
        return message + "."



import click
from flask import Flask

from tuno.server.utils.Logger import Logger, LogLevel
from tuno.shared.constraints import (
    DEFAULT_PLAYER_CAPACITY,
    MAX_PLAYER_CAPACITY,
    MIN_PLAYER_CAPACITY,
)

from .config import DEFAULT_HOST, DEFAULT_PORT


def create_app() -> Flask:

    from .models.Game import game
    from .routes import load_routes

    game.watcher_thread.start()

    app = Flask(__name__)

    blueprint = load_routes()
    app.register_blueprint(blueprint, url_prefix="/api")

    return app


@click.command("server")
@click.option(
    "--host",
    default=DEFAULT_HOST,
    show_default=True,
    help="Server host",
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=DEFAULT_PORT,
    show_default=True,
    help="Server port",
)
@click.option(
    "-c",
    "--capacity",
    type=click.IntRange(
        min=MIN_PLAYER_CAPACITY,
        max=MAX_PLAYER_CAPACITY,
    ),
    default=DEFAULT_PLAYER_CAPACITY,
    show_default=True,
    help="Initial player capacity",
)
@click.option(
    "-l",
    "--log-level",
    type=click.Choice([l.name for l in LogLevel], case_sensitive=False),
    envvar="UNO_LOG_LEVEL",
    default=LogLevel.INFO.name,
    show_default=True,
    show_envvar=True,
    show_choices=True,
    help="Log level",
)
def start_server(
    host: str,
    port: int,
    capacity: int,
    log_level: str,
) -> None:
    """Start game server."""

    Logger.level = LogLevel[log_level]

    from .models.Game import game

    game.update_rules(
        {
            "player_capacity": capacity,
        },
        operator_name="server setup",
        operator_is_player=False,
    )

    app = create_app()
    app.run(host=host, port=port)



from textual.reactive import reactive
from textual.widgets import Label

from tuno.shared.deck import BasicCardColor


class CardColorLabel(Label):

    DEFAULT_CSS = """
    CardColorLabel {
        width: 100%;
        height: 100%;
        content-align-horizontal: center;
        color: $text;
    }
    """

    data: reactive[BasicCardColor | None] = reactive(None)

    def watch_data(self, data: BasicCardColor | None) -> None:
        if data is None:
            self.styles.background = "gray 50%"
            self.update("???")
            self.tooltip = None
        else:
            self.styles.background = f"{data} 90%"
            self.update(data.title())
            self.tooltip = data.title()



from collections.abc import Mapping
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING, Protocol, runtime_checkable

from textual import log

if TYPE_CHECKING:
    from tuno.client.UnoApp import UnoApp

__handler_dir = Path(__file__).parent


@runtime_checkable
class EventHandler(Protocol):
    def __call__(self, parsed_data: object, app: "UnoApp") -> None: ...


type EventHandlerMap = Mapping[str, EventHandler]


def load_event_handler_map() -> EventHandlerMap:

    event_handler_map: dict[str, EventHandler] = {}

    for entry in __handler_dir.iterdir():

        if entry.name.startswith("_"):
            continue

        module_name: str
        if entry.is_file():
            if entry.suffix != ".py":
                continue
            module_name = entry.stem
        else:
            if not (entry / "__init__.py").exists():
                continue
            module_name = entry.name

        module = import_module(f".{module_name}", __name__)
        if not (
            hasattr(module, "handler") and isinstance(module.handler, EventHandler)
        ):
            raise RuntimeError(f"Invalid event handler module: {entry}")

        if module_name in event_handler_map:
            raise RuntimeError(f"Duplicate event handler module: {entry}")

        event_handler_map[module_name] = module.handler
        log.debug(f"Loaded event handler: {module_name}")

    return event_handler_map



from collections.abc import Generator
from datetime import timedelta
from time import monotonic, sleep
from typing import Protocol


class LoopSkipCallback(Protocol):
    def __call__(self, continuous_slow_count: int) -> None: ...


def loop(
    interval_seconds: timedelta | float,
    *,
    allow_skip: bool = False,
    on_slow: LoopSkipCallback | None = None,
) -> Generator[float]:
    """A loop helper that yields at specific interval.

    Args:
        interval_seconds (timedelta | float): Expected interval between each yield.
        allow_skip (bool, optional): Whether to allow frame skip. Defaults to False.
        on_slow (LoopSkipCallback, optional): A callback function to be called when
            the duration of some iteration exceeds expected interval. The argument
            of the callback indicates the number of continuous slow iterations.
            Defaults to None.

    Yields:
        begin_timestamp (float): The timestamp when current loop begins, returned
            by `time.monotonic()`.
    """

    if isinstance(interval_seconds, timedelta):
        interval_seconds = interval_seconds.total_seconds()

    continuous_slow_count = 0

    while True:

        timestamp_begin = monotonic()
        yield timestamp_begin

        timestamp_end = monotonic()
        if timestamp_end - timestamp_begin < interval_seconds:

            loop_duration = timestamp_end - timestamp_begin
            sleep_time = interval_seconds - loop_duration

            if sleep_time < 0:
                continuous_slow_count += 1
                if on_slow:
                    on_slow(continuous_slow_count)
                    if allow_skip:
                        timestamp_skip = monotonic()
                        sleep_time = (
                            timestamp_skip - timestamp_begin
                        ) % interval_seconds
                    else:
                        sleep_time = 0
                else:
                    if allow_skip:
                        sleep_time %= interval_seconds
                    else:
                        sleep_time = 0
            else:
                continuous_slow_count = 0

            sleep(sleep_time)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tuno.client.UnoApp import UnoApp


def handler(parsed_data: None, app: "UnoApp") -> None:
    app.notify(
        "Connection has been replaced by a new one.",
        title="Subscription Change",
        severity="warning",
    )



import json
from abc import ABC
from dataclasses import dataclass
from typing import Any, TypedDict

from tuno.shared.rules import GameRules

from .deck import BasicCardColor, Card, Deck


class ServerSentEvent(ABC):

    type: str
    data: Any = None

    def to_sse(self) -> str:
        result = f"event: {self.type}\n"
        if self.data != None:
            result += f"data: {json.dumps(self.data)}\n"
        return result + "\n"


@dataclass
class EndOfConnectionEvent(ServerSentEvent):
    """
    This event stops receiving player's subscription handler
    from processing any more messages the handler should end the connection
    right after sending this event.
    """

    type = "end_of_connection"
    data: str
    """A message that explains the end of connection."""


@dataclass
class NotificationEvent(ServerSentEvent):

    class DataType(TypedDict):
        title: str
        message: str

    type = "notification"
    data: DataType


@dataclass
class SubscriptionChangeEvent(ServerSentEvent):
    type = "subscription_change"


@dataclass
class GameStateEvent(ServerSentEvent):

    class PlayerDataType(TypedDict):
        name: str
        connected: bool
        card_count: int

    class DataType(TypedDict):
        started: bool
        rules: GameRules
        draw_pile_size: int
        discard_pile_size: int
        players: "list[GameStateEvent.PlayerDataType]"
        current_player_index: int
        direction: int
        lead_card: Card | None
        lead_color: BasicCardColor | None
        draw_counter: int
        skip_counter: int

    type = "game_state"
    data: DataType


@dataclass
class CardsEvent(ServerSentEvent):
    type = "cards"
    data: Deck



from textual.app import ComposeResult
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Label, LoadingIndicator


class LoadingScreen(ModalScreen[object]):

    CSS_PATH = "styles.tcss"

    message: str

    def __init__(self, message: str = "Loading...") -> None:
        super().__init__()
        self.message = message

    def compose(self) -> ComposeResult:
        yield Vertical(
            Label(self.message, id="loading-message"),
            LoadingIndicator(),
            id="loading-container",
        )



from enum import IntEnum
from functools import partialmethod
from threading import RLock
from time import strftime

import click


class LogLevel(IntEnum):
    DEBUG = 1
    INFO = 2
    WARN = 3
    ERROR = 4


class Logger:

    name: str
    level: LogLevel = LogLevel.INFO
    lock = RLock()  # cross-thread io lock

    def __init__(self, name: str) -> None:
        self.name = name

    def __log(
        self,
        content: str,
        *,
        level: LogLevel,
        label: str,
        fg: str,
        bg: str,
    ) -> None:
        if level < self.level:
            return
        timestamp = strftime("%Y-%m-%d %H:%M:%S")
        message = (
            click.style(
                f"[{timestamp}] {label:>5s} ({self.name})",
                fg=fg,
                bg=bg,
            )
            + " "
            + content
        )
        with self.lock:
            click.echo(message)

    debug = partialmethod(
        __log,
        level=LogLevel.DEBUG,
        label="DEBUG",
        fg="blue",
        bg="black",
    )

    info = partialmethod(
        __log,
        level=LogLevel.INFO,
        label="INFO",
        fg="green",
        bg="black",
    )

    warn = partialmethod(
        __log,
        level=LogLevel.WARN,
        label="WARN",
        fg="yellow",
        bg="black",
    )

    error = partialmethod(
        __log,
        level=LogLevel.ERROR,
        label="ERROR",
        fg="red",
        bg="yellow",
    )



from typing import TYPE_CHECKING

from tuno.shared.deck import Deck

if TYPE_CHECKING:
    from tuno.client.UnoApp import UnoApp


def handler(parsed_data: Deck, app: "UnoApp") -> None:

    assert app.client is not None
    app.client.cards = parsed_data

    app.post_message(app.CardsUpdate(parsed_data))



from random import shuffle
from typing import Final

from ...shared.deck import (
    Deck,
    FunctionCard,
    NumberCard,
    WildCard,
    basic_card_colors,
    function_card_effects,
    wild_card_color,
    wild_card_effects,
)

INITIAL_DECK_SIZE: Final[int] = 108


def create_deck() -> Deck:
    """Create and return a non-shuffled deck."""

    deck: Deck = []

    ids = [str(x) for x in range(INITIAL_DECK_SIZE)]
    shuffle(ids)

    # number cards
    for color in basic_card_colors:
        for number in range(10):
            deck.append(
                NumberCard(
                    id=ids.pop(),
                    color=color,
                    type="number",
                    number=number,
                )
            )
            if number != 0:
                deck.append(
                    NumberCard(
                        id=ids.pop(),
                        color=color,
                        type="number",
                        number=number,
                    )
                )

    # function cards
    for color in basic_card_colors:
        for function_card_effect in function_card_effects:
            for _ in range(2):
                deck.append(
                    FunctionCard(
                        id=ids.pop(),
                        color=color,
                        type="function",
                        effect=function_card_effect,
                    )
                )

    # wild cards
    for _ in range(4):
        for wild_card_effect in wild_card_effects:
            deck.append(
                WildCard(
                    id=ids.pop(),
                    color=wild_card_color,
                    type="wild",
                    effect=wild_card_effect,
                )
            )

    assert len(ids) == 0
    assert len(deck) == INITIAL_DECK_SIZE

    return deck



from typing import cast

from textual import work
from textual.app import ComposeResult
from textual.containers import HorizontalScroll
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import Button, Footer, Header

from tuno.client.components.CardsScreen import CardsScreen
from tuno.client.components.Players import Players
from tuno.client.components.RulesScreen import RulesScreen
from tuno.client.utils.LoadingContext import LoadingContext
from tuno.shared.deck import Deck
from tuno.shared.sse_events import GameStateEvent

from .Sidebar import Sidebar


class InGameScreen(Screen[object]):

    TITLE = "UNO"
    CSS_PATH = "styles.tcss"
    BINDINGS = [
        ("ctrl+t", "play", "Play"),
        ("ctrl+r", "show_rules", "Rules"),
    ]

    game_state: reactive[GameStateEvent.DataType | None] = reactive(None)
    cards: reactive[Deck] = reactive([])

    def compose(self) -> ComposeResult:

        actions_container = HorizontalScroll(
            Button(
                "Play",
                id="action-show-play",
                variant="primary",
                action="screen.play",
            ),
            Button(
                "Rules",
                id="action-show-rules",
                variant="warning",
                action="screen.show_rules",
            ),
            Button(
                "Stop",
                id="action-stop-game",
                variant="error",
                action="screen.stop_game",
            ),
            id="actions",
        )
        actions_container.border_title = "Actions"

        yield Header(show_clock=True)
        yield Sidebar().data_bind(InGameScreen.game_state)
        yield Players().data_bind(InGameScreen.game_state)
        yield actions_container
        yield Footer()

    def watch_game_state(self, game_state: GameStateEvent.DataType | None) -> None:

        from tuno.client.UnoApp import UnoApp

        app = cast(UnoApp, self.app)
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None
        self.sub_title = client.get_connection_display()

    def action_play(self) -> None:
        self.app.push_screen(CardsScreen())

    def action_show_rules(self) -> None:
        self.app.push_screen(RulesScreen(readonly=True))

    @work(thread=True)
    def action_stop_game(self) -> None:

        from tuno.client.UnoApp import UnoApp

        app = cast(UnoApp, self.app)
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None

        with LoadingContext("Stopping game...", app=app):
            client.stop_game()



import re
from typing import Final

PLAYER_NAME_PATTERN: Final = re.compile(r"^[A-Za-z0-9_-]{1,20}$")

MIN_PLAYER_CAPACITY: Final = 2
MAX_PLAYER_CAPACITY: Final = 20
DEFAULT_PLAYER_CAPACITY: Final = 8
assert MIN_PLAYER_CAPACITY <= DEFAULT_PLAYER_CAPACITY <= MAX_PLAYER_CAPACITY

MIN_INITIAL_HAND_SIZE: Final = 2
MAX_INITIAL_HAND_SIZE: Final = 16
DEFAULT_INITIAL_HAND_SIZE: Final = 7
assert MIN_INITIAL_HAND_SIZE <= DEFAULT_INITIAL_HAND_SIZE <= MAX_INITIAL_HAND_SIZE



from flask import Blueprint, request
from flask.typing import ResponseReturnValue

from tuno.server.utils.checkers import check_player_name
from tuno.server.utils.Logger import Logger

logger = Logger(__name__)


def setup(blueprint: Blueprint) -> None:

    @blueprint.delete("/<target_player_name>")
    def kick_out(target_player_name: str) -> ResponseReturnValue:

        check_player_name(target_player_name)

        operator_name = request.args.get("operator", "")
        if operator_name:
            check_player_name(operator_name)
        else:
            operator_name = "(anonymous)"

        from tuno.server.models.Game import game

        game.kick_out_player(
            target_name=target_player_name,
            operator_name=operator_name,
            operator_is_player=True,
        )

        return ("Player kicked out.", 200)



from importlib import import_module
from pathlib import Path

from flask import Blueprint

from tuno.server.exceptions import ApiException
from tuno.server.utils.Logger import Logger

__logger = Logger(__name__)
__routes_dir = Path(__file__).parent


def __load_routes(blueprint_name: str, folder: Path) -> Blueprint:

    blueprint = Blueprint(blueprint_name, __name__)

    for entry in folder.iterdir():

        if entry.name.startswith("_"):
            continue

        if entry.is_file() or (entry / "__init__.py").exists():

            if entry.is_file() and entry.suffix.lower() != ".py":
                continue

            route_name = ".".join(
                entry.relative_to(__routes_dir).parts,
            )[:-3]
            module = import_module(f".{route_name}", __name__)

            if hasattr(module, "setup"):
                module.setup(blueprint)
                __logger.debug(f'Loaded route "{route_name}".')
            else:
                raise RuntimeError(
                    f'Route "{route_name}" does not have a setup function!'
                )

        else:

            sub_blueprint = __load_routes(entry.name, entry)
            blueprint.register_blueprint(
                sub_blueprint,
                url_prefix=f"/{entry.name}",
            )

    return blueprint


def load_routes() -> Blueprint:

    blueprint = __load_routes("backend", __routes_dir)

    @blueprint.errorhandler(ApiException)
    def handle_api_exception(exception: ApiException) -> tuple[str, int]:
        __logger.warn(f"ApiException({exception.http_code}): {exception.message}")
        return exception.message, exception.http_code

    return blueprint



from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, Literal, Self

from tuno.client.components.LoadingScreen import LoadingScreen

if TYPE_CHECKING:
    from tuno.client.UnoApp import UnoApp


class LoadingContext(AbstractContextManager["LoadingContext"]):

    message: str
    app: "UnoApp"
    loading_screen: LoadingScreen | None

    def __init__(self, message: str, *, app: "UnoApp") -> None:
        self.message = message
        self.app = app
        self.loading_screen = None

    def __enter__(self) -> Self:
        if self.loading_screen is not None:
            raise RuntimeError("LoadingContext entered more than once")
        self.loading_screen = LoadingScreen(self.message)
        self.app.call_from_thread(
            self.app.push_screen,
            self.loading_screen,
        )
        return self

    def __exit__(
        self,
        exception_type: type[BaseException] | None,
        exception: BaseException | None,
        traceback: object,
    ) -> Literal[False]:
        if self.loading_screen is None:
            raise RuntimeError("LoadingContext exited without entering")
        self.app.call_from_thread(self.loading_screen.dismiss)
        self.loading_screen = None
        return False  # propagate exceptions



from typing import cast

from textual import work
from textual.app import ComposeResult
from textual.containers import Horizontal, HorizontalScroll, Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Button, Footer, Header, Label, Select

from tuno.client.components.CardLabel import CardLabel
from tuno.client.components.CheckboxContainer import CheckboxContainer
from tuno.client.utils.LoadingContext import LoadingContext
from tuno.shared.deck import Card, basic_card_colors


class CardCheckbox(CheckboxContainer):

    card_data: Card

    def __init__(self, card_data: Card) -> None:
        super().__init__(CardLabel())
        self.card_data = card_data

    def on_mount(self) -> None:
        self.query_exactly_one(CardLabel).data = self.card_data


class CardsScreen(ModalScreen[object]):

    TITLE = "UNO"
    CSS_PATH = "styles.tcss"
    BINDINGS = [
        ("ctrl+s", "screen.submit", "Submit"),
        ("escape", "screen.dismiss", "Cancel"),
    ]

    def compose(self) -> ComposeResult:

        from tuno.client.UnoApp import UnoApp

        app = self.app
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None

        yield Header(show_clock=True)
        with Vertical(id="cards-window"):
            yield Label("Cards to play: (Submit none to pass.)", id="cards-title")
            yield VerticalScroll(id="cards-container")
            with Horizontal(id="cards-color-select-container"):
                yield Label(
                    "Change color: ",
                    id="cards-color-select-label",
                )
                yield Select[str](
                    ((color.title(), color) for color in basic_card_colors),
                    id="cards-color-select",
                    prompt="(none)",
                    tooltip="Color to change to.",
                )
            with HorizontalScroll(id="cards-actions"):
                yield Button(
                    "Submit",
                    id="rules-submit",
                    variant="primary",
                    action="screen.submit",
                )
                yield Button(
                    "Reset",
                    variant="warning",
                    action="screen.reset",
                )
                yield Button(
                    "Cancel",
                    variant="error",
                    action="screen.dismiss",
                )
        yield Footer()

    def on_mount(self) -> None:

        from tuno.client.UnoApp import UnoApp

        app = cast(UnoApp, self.app)
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None

        self.sub_title = client.get_connection_display()

        cards_container = self.query_exactly_one("#cards-container", VerticalScroll)
        for card in client.cards:
            cards_container.mount(CardCheckbox(card))

    @work(thread=True)
    def action_submit(self) -> None:

        from tuno.client.UnoApp import UnoApp

        app = self.app
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None
        assert client.game_state is not None

        selected_cards = [
            card_widget.card_data
            for card_widget in self.query(CardCheckbox)
            if card_widget.checked
        ]

        selected_card_ids = [card["id"] for card in selected_cards]
        color_select_value = self.query_exactly_one(Select).value
        with LoadingContext("Playing...", app=app):
            client.play(
                selected_card_ids,
                color=(
                    None
                    if color_select_value is Select.BLANK
                    else cast(str, color_select_value)
                ),
            )
            self.app.call_from_thread(self.dismiss)

    def action_reset(self) -> None:

        from tuno.client.UnoApp import UnoApp

        app = self.app
        assert isinstance(app, UnoApp)

        client = app.client
        assert client is not None
        assert client.game_state is not None

        for card_checkbox in self.query(CardCheckbox):
            card_checkbox.checked = False

        self.query_exactly_one("#cards-color-select", Select).value = Select.BLANK



from datetime import timedelta
from typing import Final

# -- Server Config --
DEFAULT_HOST: Final = "0.0.0.0"
DEFAULT_PORT: Final = 5000

# -- Connection Config --
SUBSCRIPTION_TOKEN_BYTES: Final = 4  # each byte becomes 2 hex digits
SUBSCRIPTION_LOOP_INTERVAL = timedelta(milliseconds=200)
SUBSCRIPTION_LOOP_SLOW_THRESHOLD: Final = 3
MAX_MESSAGES_PER_LOOP: Final = 5
HEARTBEAT_GAP = timedelta(seconds=2)
PLAYER_TIMEOUT = timedelta(seconds=5)

# -- Game Config --
GAME_WATCHER_INTERVAL = timedelta(seconds=2)
GAME_WATCHER_SKIP_THRESHOLD: Final = 3

# -- Player Config --
PLAYER_MESSAGE_QUEUE_SIZE: Final = 20
